import glob
import os
import subprocess
from asyncio import AbstractEventLoop
from pathlib import Path
from typing import (
    Dict,
    Text,
    Any,
    Optional,
    List,
    Tuple,
    Generator,
    Callable,
    Coroutine,
)
from uuid import uuid4

import pytest
import time

from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempdirFactory
from pytest_sanic.utils import TestClient

from rasax.community import utils
from rasax.community.api.blueprints import git
from rasax.community.constants import (
    COMMUNITY_USERNAME,
    REQUEST_DB_SESSION_KEY,
    CONFIG_METRICS_ENABLED,
    CONFIG_METRICS_USER,
)
from sanic import Sanic, Blueprint
from sanic.testing import SanicTestClient
from rasax.community.constants import DEFAULT_RASA_ENVIRONMENT
from sqlalchemy.orm import Session, sessionmaker

import rasax.community.config as rasa_x_config
import rasax.community.jwt
import rasax.community.sql_migrations as migrations
from rasa.core.events import UserUttered
from rasa.utils.endpoints import EndpointConfig
from rasax.community.config import team_name
from rasax.community.services.analytics_service import AnalyticsService
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.evaluation_service import EvaluationService
from rasax.community.services.event_service import EventService
from rasax.community.services.feature_service import FeatureService
from rasax.community.services.git_service import GitService
from rasax.community.services.intent_service import IntentService
from rasax.community.services.logs_service import LogsService
from rasax.community.services.model_service import ModelService
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.role_service import RoleService
from rasax.community.services.settings_service import SettingsService
from rasax.community.services.stack_service import StackService
from rasax.community.services.story_service import StoryService
from rasax.community.services.user_goal_service import UserGoalService
from rasax.community.services.user_service import UserService, ADMIN

UUID = uuid4().hex

TEST_USER = "some_user"
TEST_PASSWORD = "password"
NLU_DATA_PATH = "tests/unit/demo-rasa.md"
STORY_PATH = "tests/unit/stories.md"
TEST_STACK_URL = "http://stack-service.com"
TEST_PROJECT = "project"


def session_maker_for_tests() -> sessionmaker:
    from rasax.community.database.utils import create_session_maker

    in_memory_sqlite_database = "sqlite://"

    # turn on local mode so that the session is initialised with SQLite settings
    rasa_x_config.LOCAL_MODE = True
    session_maker = create_session_maker(in_memory_sqlite_database)

    # turn it off again
    rasa_x_config.LOCAL_MODE = False

    return session_maker


@pytest.fixture()
def session_without_migration() -> Generator[sessionmaker, None, None]:
    import rasax.community.utils

    session_maker = session_maker_for_tests()
    # Allow multiple user creation for tests
    rasax.community.utils.is_enterprise_installed = lambda: True

    session = session_maker()
    yield session
    session.commit()
    session.close()


@pytest.fixture(scope="module")
def session() -> Generator[Session, None, None]:
    import rasax.community.utils

    session_maker = session_maker_for_tests()
    session = session_maker()

    # Allow multiple user creation for tests
    rasax.community.utils.is_enterprise_installed = lambda: True

    migrations.run_migrations(session)

    # Insert general test user which can be used in later tests
    insert_test_user(session, TEST_USER)

    # Insert default community user
    insert_test_user(session, COMMUNITY_USERNAME)

    session = session_maker()
    yield session
    session.commit()
    session.close()


@pytest.fixture()
def app(session: Session, monkeypatch: MonkeyPatch) -> Sanic:
    from rasax.community.api import app

    def test_session() -> Session:
        return session

    # Use our test database session
    import rasax.community.database.utils

    monkeypatch.setattr(
        rasax.community.database.utils, "create_session_maker", lambda _: test_session
    )

    # Don't include frontend blueprint in unit tests (use empty blueprint instead)
    from rasax.community.api.blueprints import interface

    monkeypatch.setattr(interface, "blueprint", lambda: Blueprint("interface"))

    # Don't load frontend
    _app = app.configure_app()

    # Set up authentication
    rasax.community.jwt.initialise_jwt_keys()
    app.initialize_app(_app)

    return _app


@pytest.fixture
def test_client(
    loop: AbstractEventLoop,
    app: Sanic,
    sanic_client: Callable[[Sanic], Coroutine],
    monkeypatch: MonkeyPatch,
    tmpdir: Path,
) -> TestClient:
    # Patch git service to use a temporary directory by default
    monkeypatch.setattr(
        git,
        "_git",
        lambda request, project, repository_id=None: GitService(
            request[REQUEST_DB_SESSION_KEY], tmpdir, project, repository_id
        ),
    )

    return loop.run_until_complete(sanic_client(app))


@pytest.fixture()
async def auth_header(test_client: SanicTestClient) -> Dict[Text, Text]:
    response = await test_client.post(
        "/api/auth", json={"username": TEST_USER, "password": TEST_PASSWORD}
    )
    response_body = await response.json()
    return {"Authorization": "Bearer " + response_body["access_token"]}


@pytest.fixture(scope="module")
def data_service(session: Session) -> DataService:
    return DataService(session)


@pytest.fixture(scope="module")
def role_service(session) -> RoleService:
    return RoleService(session)


@pytest.fixture(scope="module")
def user_service(session) -> UserService:
    return UserService(session)


@pytest.fixture(scope="module")
def user_goal_service(session) -> UserGoalService:
    return UserGoalService(session)


@pytest.fixture(scope="module")
def settings_service(session) -> SettingsService:
    return SettingsService(session)


@pytest.fixture(scope="module")
def domain_service(session) -> DomainService:
    return DomainService(session)


@pytest.fixture(scope="module")
def model_service(session, tmpdir_factory) -> ModelService:
    model_directory = tmpdir_factory.mktemp("models")

    return ModelService(model_directory, session, DEFAULT_RASA_ENVIRONMENT)


@pytest.fixture(scope="module")
def logs_service(session):
    return LogsService(session)


@pytest.fixture(scope="module")
def intent_service(session):
    return IntentService(session)


@pytest.fixture(scope="module")
def event_service(session):
    return EventService(session)


@pytest.fixture(scope="module")
def analytics_service(session):
    return AnalyticsService(session)


@pytest.fixture(scope="module")
def nlg_service(session):
    return NlgService(session)


@pytest.fixture(scope="module")
def story_service(session):
    return StoryService(session)


@pytest.fixture(scope="module")
def feature_service(session):
    return FeatureService(session)


@pytest.fixture(scope="module")
def evaluation_service(session):
    return EvaluationService(None, session)


@pytest.fixture(scope="module")
def stack_service(session) -> StackService:
    # TODO: create an actual Rasa service that runs as a background process
    _config = EndpointConfig(TEST_STACK_URL)
    return StackService(
        _config,
        DataService(session),
        StoryService(session),
        DomainService(session),
        SettingsService(session),
    )


@pytest.fixture(scope="module")
def user(session) -> Dict[Text, Any]:
    user_service = UserService(session)
    user_service.create_user("test user", TEST_PASSWORD, rasa_x_config.team_name, ADMIN)
    return user_service.fetch_user("test user")


@pytest.fixture(scope="session")
def config():
    return """
    language: "en"
    pipeline: "spacy_sklearn"
    """


def _user_event(
    sender_id: Text,
    timestamp: float = 0,
    intent: Text = "test",
    project: Text = "default",
    text: Text = "hello",
    entities: Optional[List] = None,
) -> Dict[Text, Any]:
    return {
        "sender_id": sender_id,
        "timestamp": timestamp,
        "event": UserUttered.type_name,
        "text": text,
        "input_channel": "rest",
        "parse_data": {
            "intent": {"name": intent},
            "entities": entities or [],
            "project": project,
            "text": text,
        },
    }


def _action_event(
    sender_id,
    timestamp=0,
    action_name="action",
    policy="policy",
    confidence=1.0,
    event_type="action",
):
    return {
        "sender_id": sender_id,
        "policy": policy,
        "confidence": confidence,
        "event": event_type,
        "name": action_name,
        "timestamp": timestamp,
    }


def _bot_event(sender_id, timestamp):
    return _action_event(sender_id, timestamp, event_type="bot")


def insert_test_model(
    project_name: Text,
    model_name: Text,
    model_service: ModelService,
    stored_path: Optional[Text] = None,
    training_time: float = time.time(),
    version: Text = "1.0",
):
    expected = {
        "project": project_name,
        "model_name": model_name,
        "stored_path": stored_path or uuid4().hex,
        "model_hash": "hash",
        "_version": version,
        "training_time": training_time,
    }

    # noinspection PyProtectedMember
    model_service._save_model_data(**expected)

    return expected


def insert_test_user(
    session: Session,
    username: Text,
    team: Text = rasa_x_config.team_name,
    role: Text = ADMIN,
):
    user_service = UserService(session)
    user_service.create_user(username, TEST_PASSWORD, team, role)
    session.commit()  # commit session in case services using another session rely on it


def user_dict(username=TEST_USER):
    return {"username": username}


@pytest.fixture
def new_project(settings_service: SettingsService):
    project = "{}-{}".format(TEST_PROJECT, uuid4().hex)
    settings_service.init_project(team_name, project)
    settings_service.commit()
    return project


def generate_key_pair(tmpdir_factory: TempdirFactory, n: int = 0) -> Tuple[Text, Text]:
    # noinspection PyProtectedMember
    private_key, public_key = rasax.community.jwt._generate_rsa_key_pair()

    key_dir = tmpdir_factory.mktemp("keys")
    private_key_path = str(key_dir / f"key_{n}.pem")
    public_key_path = str(key_dir / f"public_key_{n}.pem")

    with open(private_key_path, "wb") as f:
        f.write(private_key)

    with open(public_key_path, "wb") as f:
        f.write(public_key)

    return private_key_path, public_key_path


@pytest.fixture(scope="session")
def _train_rasa_model(tmpdir_factory: TempdirFactory) -> Text:
    # creates a scaffold project and trains a minimal rasa model

    from rasa.cli import scaffold
    from rasa.constants import DEFAULT_CONFIG_PATH
    import rasa.nlu.utils as rasa_nlu_utils

    # create a rasa scaffold in temporary directory
    project_dir = tmpdir_factory.mktemp("rasa-project")
    scaffold.create_initial_project(str(project_dir))

    # overwrite config with minimal config to speed up training
    minimal_config = """
    language: "en"
    pipeline:
      - name: "WhitespaceTokenizer"
    policies:
      - name: MemoizationPolicy
    """

    rasa_nlu_utils.write_to_file(
        os.path.join(project_dir, DEFAULT_CONFIG_PATH), minimal_config
    )

    subprocess.call(["rasa", "train"], cwd=project_dir)

    model_name = os.listdir(os.path.join(project_dir, "models"))[0]

    return os.path.join(project_dir, "models", model_name)


def _list_files_in_dir(dirname: Text) -> List:
    # return a list of all files in directory `dirname`
    return [f for f in glob.glob(os.path.join(dirname, "*"))]


@pytest.fixture(scope="function")
def rasa_model_path(
    _train_rasa_model: Text, model_service: ModelService
) -> Generator[Text, None, None]:
    model_files_before_fixture = _list_files_in_dir(model_service.model_directory)

    yield _train_rasa_model

    model_files_after_fixture = _list_files_in_dir(model_service.model_directory)

    # remove all new files in model directory created by the test function
    for f in set(model_files_after_fixture) - set(model_files_before_fixture):
        os.remove(f)


@pytest.fixture
def enable_metrics(monkeypatch: MonkeyPatch) -> None:
    from tests.unit.test_metrics import METRICS_TEST_USER
    from rasax.community import config

    new_config = {CONFIG_METRICS_ENABLED: True, CONFIG_METRICS_USER: METRICS_TEST_USER}
    monkeypatch.setattr(config, "metrics_collection_config", new_config)


@pytest.fixture
def metrics_events(enable_metrics: None, monkeypatch: MonkeyPatch) -> List:
    from rasax.community import metrics

    events = []

    def fn(_, event_name, properties=None, context=None):
        if properties:
            properties.pop("metrics_id", None)

        events.append(
            {"event": event_name, "properties": properties, "context": context}
        )

    monkeypatch.setattr(metrics, "_track_internal", fn)
    return events
