import logging
import os
import stat
import time
import uuid
from pathlib import Path
from typing import Text, Dict, List

import pytest
from _pytest.logging import LogCaptureFixture
from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempdirFactory
from rasa.utils.io import DEFAULT_ENCODING
from rasax.community.database.admin import GitRepository
from sqlalchemy.orm import Session

from rasax.community.services.git_service import (
    GitService,
    SSH_FILES_DIRECTORY,
    DEFAULT_TARGET_BRANCH,
)
from rasax.community import utils
from rasax.community.services.data_service import DataService

# The SSH key to access this repository is stored in the travis settings
SSH_KEY = os.environ.get("SSH_KEY_FOR_GIT_TEST_REPOSITORY")

if SSH_KEY:
    # Make sure newlines are not double escaped
    SSH_KEY = SSH_KEY.encode(DEFAULT_ENCODING).decode("unicode_escape")

RASA_REMOTE_REPOSITORY = {
    "repository_url": "git@github.com:RasaHQ/git-test.git",
    "ssh_key": SSH_KEY,
}


@pytest.fixture()
def bare_git_service(
    session: Session, tmpdir: Path, monkeypatch: MonkeyPatch
) -> GitService:
    """This returns `GitService` instance where the repository is not yet cloned."""

    git_service = GitService(session, tmpdir)
    # Turn off permission validation for initial setup to increase test speed
    with monkeypatch.context() as m:
        m.setattr(git_service, "has_required_remote_permission", lambda _, __: True)
        _ = git_service.save_repository(RASA_REMOTE_REPOSITORY)

    return git_service


@pytest.fixture(scope="module")
def cloned_directory(session: Session, tmpdir_factory: TempdirFactory) -> Path:
    tmpdir = tmpdir_factory.mktemp("base")
    git_service = GitService(session, tmpdir)
    _ = git_service.save_repository(RASA_REMOTE_REPOSITORY)
    git_service.clone()
    return git_service.repository_path()


@pytest.fixture()
def git_service_with_credentials_validation(
    bare_git_service: GitService, cloned_directory: Path
) -> GitService:
    """Returns a `GitService` instance where the repository was already cloned."""

    import shutil

    repository_info = bare_git_service.get_most_recent_repository()
    path = bare_git_service.repository_path(repository_info.id)

    # Don't clone each time, rather reset the directory by removing everything and
    # copying in the original state
    shutil.rmtree(path)
    shutil.copytree(cloned_directory, path)
    bare_git_service.set_repository(repository_info.id)

    return bare_git_service


@pytest.fixture()
def git_service(
    git_service_with_credentials_validation: GitService, monkeypatch: MonkeyPatch
) -> GitService:
    """Returns a `GitService` instance which skips the `has_required_remote_permission`
    check."""

    monkeypatch.setattr(
        git_service_with_credentials_validation,
        "has_required_remote_permission",
        lambda _, __: True,
    )

    return git_service_with_credentials_validation


def test_git_directory_initialization(git_service: GitService):
    assert (git_service.repository_path() / ".git").exists()


def test_get_current_branch(git_service: GitService):
    assert git_service.get_current_branch() == DEFAULT_TARGET_BRANCH


def test_create_branch(git_service: GitService):
    branch_name = "test-create-branch"
    git_service.create_branch(branch_name)

    assert branch_name in [branch.name for branch in git_service._repository.branches]


def test_create_branch_if_it_already_exists(git_service: GitService):
    # This should not give an exception
    git_service.create_branch(DEFAULT_TARGET_BRANCH)


def test_checkout_branch(git_service: GitService):
    branch_name = "test-checkout-branch"
    git_service.create_branch(branch_name)
    git_service.checkout_branch(branch_name)

    assert git_service.get_current_branch() == branch_name


def test_checkout_branch_if_dirty_and_force(git_service: GitService):
    branch_name = "test-checkout-branch"
    git_service.create_branch(branch_name)

    # Make it dirty 🚮
    new_untracked_file = git_service.repository_path() / "folder" / "new-file.txt"
    utils.write_text_to_file(new_untracked_file, "content")
    tracked_file = git_service.repository_path() / "data" / "nlu.md"
    utils.write_text_to_file(tracked_file, "content")

    git_service.checkout_branch(branch_name, force=True)

    assert git_service.get_current_branch() == branch_name
    assert not new_untracked_file.exists()
    assert not git_service.is_dirty()


def test_checkout_branch_if_branch_does_not_exist(git_service: GitService):
    with pytest.raises(ValueError):
        git_service.checkout_branch("not-existing-branch-name")


def test_is_dirty(git_service: GitService):
    assert not git_service.is_dirty()

    utils.write_text_to_file(git_service.repository_path() / "new-file.txt", "content")

    assert git_service.is_dirty()

    assert git_service.get_repository_status()["are_there_local_changes"]


def test_commit_to_target_branch(git_service: GitService):
    # Add local changes
    new_file = git_service.repository_path() / "nlu.md"
    utils.write_text_to_file(new_file, "##intent")

    # Commit changes
    commit_message = "test test"
    branch = git_service.commit_to_target_branch(commit_message)

    assert branch == DEFAULT_TARGET_BRANCH
    assert git_service.get_current_branch() == DEFAULT_TARGET_BRANCH
    # Check that no new branch was created
    assert [branch.name for branch in git_service._repository.branches] == [
        DEFAULT_TARGET_BRANCH
    ]
    assert git_service._repository.commit().message == commit_message


def test_commit_to_new_branch(git_service: GitService):
    # Add local changes
    new_file = git_service.repository_path() / "nlu.md"
    utils.write_text_to_file(new_file, "##intent")

    # Commit changes to new branch
    commit_message = "test test"
    new_branch_name = "new-branch"
    git_service.commit_to(new_branch_name, commit_message)

    # Assert that we switched back to target branch after commit
    assert git_service.get_current_branch() == DEFAULT_TARGET_BRANCH
    # Assert a new branch was created
    assert new_branch_name in [
        branch.name for branch in git_service._repository.branches
    ]
    # Assert new branch contains changes
    assert (
        git_service._repository.branches[new_branch_name].commit.message
        == commit_message
    )


def test_clone(git_service: GitService):
    assert git_service.get_current_branch() == DEFAULT_TARGET_BRANCH
    assert "README.md" in os.listdir(git_service.repository_path())


def test_clone_in_existing_git_directory(git_service: GitService):
    with pytest.raises(ValueError):
        # clone after it was already cloned
        git_service.clone()


def test_push(git_service: GitService, metrics_events: List):
    # Add local changes
    new_file = git_service.repository_path() / "nlu.md"
    utils.write_text_to_file(new_file, "##intent")

    # Push changes to remote
    new_branch_name = str(uuid.uuid4())
    git_service.commit_and_push_changes_to(new_branch_name)

    # Assert changes were pushed
    remote_branch_name = f"origin/{new_branch_name}"
    assert remote_branch_name in [
        ref.name for ref in git_service._repository.remote().refs
    ]

    # Assert push was tracked
    assert metrics_events

    # Delete new branch from remote
    git_service.delete_remote_branch(new_branch_name)


def test_push_changes_to_protected_branch(
    git_service: GitService, metrics_events: List
):
    # Branch in test repository which is marked as `protected`
    protected_branch_name = "protected-branch"

    # Update repository to use protected branch as target branch
    repository = RASA_REMOTE_REPOSITORY.copy()
    repository["target_branch"] = protected_branch_name
    git_service.update_repository(repository)

    # Add local changes
    new_file = git_service.repository_path() / "nlu.md"
    utils.write_text_to_file(new_file, "## greet")

    # Commit and push local changes
    push_result = git_service.commit_and_push_changes_to(protected_branch_name)
    pushed_branch_name = push_result["remote_branch"]

    assert pushed_branch_name != protected_branch_name
    assert push_result["committed_to_target_branch"] is False
    assert git_service.get_current_branch() == protected_branch_name
    assert git_service.is_remote_branch_ahead(pushed_branch_name)

    # Assert push was tracked
    assert metrics_events

    # Delete new branch from remote
    git_service.delete_remote_branch(pushed_branch_name)


def test_is_remote_branch_ahead(git_service: GitService):
    assert not git_service.is_remote_branch_ahead()

    # Add local changes
    new_file = git_service.repository_path() / "nlu.md"
    utils.write_text_to_file(new_file, "##intent")

    # Commit and push changes to new remote branch
    new_branch_name = str(uuid.uuid4())
    git_service.commit_to(new_branch_name, "test_push")
    git_service._push(new_branch_name)

    # Get current time so that we can check if `is_remote_branch_ahead` actually
    # fetched latest remote changes
    now = time.time()

    # Assert pushed remote branch is one ahead of the current target branch
    assert git_service.is_remote_branch_ahead(new_branch_name)
    assert git_service._get_latest_fetch_time() >= now

    # Delete new branch from remote
    git_service.delete_remote_branch(new_branch_name)


def test_merge_remote_branch(git_service: GitService):
    new_file = git_service.repository_path() / "nlu.md"
    utils.write_text_to_file(new_file, "##intent")

    new_branch_name = str(uuid.uuid4())
    git_service.commit_to_new_branch(new_branch_name, "test_push")
    git_service._push(new_branch_name)

    git_service.checkout_branch(new_branch_name)
    # Nothing should happen
    git_service.merge_latest_remote_branch()

    # Set local branch 1 commit back so we have sth to pull
    git_service._repository.git.reset("--hard", DEFAULT_TARGET_BRANCH)

    assert git_service.is_remote_branch_ahead(new_branch_name)
    git_service.merge_latest_remote_branch()
    assert not git_service.is_remote_branch_ahead(new_branch_name)

    git_service.delete_remote_branch(new_branch_name)


def test_not_is_git_directory(temp_dir: Path):
    assert not GitService._is_git_directory(temp_dir)


def test_is_git_directory(git_service: GitService):
    assert GitService._is_git_directory(git_service.repository_path())


async def test_synchronize_without_any_repositories(
    session: Session, bare_git_service: GitService, caplog: LogCaptureFixture
):
    # Delete existing repositories from previous tests
    session.query(GitRepository).delete()
    session.commit()

    with caplog.at_level(logging.DEBUG):
        await bare_git_service.synchronize_project()

    assert any(
        [
            "Skip synchronizing with Git since no credentials were given."
            in record.message
            for record in caplog.records
        ]
    )


async def test_synchronize_project_when_repository_was_not_cloned_before(
    bare_git_service: GitService, data_service: DataService
):
    training_data = data_service.get_nlu_training_data_object()
    assert training_data.is_empty()
    # This should clone the repository and inject the data
    await bare_git_service.synchronize_project()

    # Assert repository was cloned
    assert GitService._is_git_directory(bare_git_service.repository_path())
    # Assert data was injected
    training_data = data_service.get_nlu_training_data_object()
    assert not training_data.is_empty()


async def test_synchronize_project_with_already_cloned_repository(
    git_service: GitService, data_service: DataService
):
    # delete training data
    data_service.delete_data()

    # Set local branch 1 commit back so we have sth to pull
    _go_n_commits_back(git_service)

    await git_service.synchronize_project()

    training_data = data_service.get_nlu_training_data_object()
    assert training_data.training_examples


def _go_n_commits_back(
    git_service: GitService, number_commits_to_go_back: int = 1
) -> None:
    git_service._repository.git.reset("--hard", f"HEAD~{number_commits_to_go_back}")


async def test_synchronize_project_if_target_branch_not_ahead(
    git_service: GitService, caplog: LogCaptureFixture
):
    with caplog.at_level(logging.DEBUG):
        await git_service.synchronize_project()

    assert any(
        ["not contain new changes" in record.message for record in caplog.records]
    )


async def test_synchronize_project_with_force_injection(
    git_service: GitService, caplog: LogCaptureFixture
):
    with caplog.at_level(logging.DEBUG):
        await git_service.synchronize_project(force_data_injection=True)

    assert any(
        ["Data injection was forced" in record.message for record in caplog.records]
    )


async def test_synchronize_project_if_working_directory_is_dirty(
    git_service: GitService, caplog: LogCaptureFixture
):
    # Set local branch 1 commit back so we have sth to pull
    _go_n_commits_back(git_service)

    # Make it dirty 🚮
    utils.write_text_to_file(git_service.repository_path() / "new-file.txt", "test")
    with caplog.at_level(logging.DEBUG):
        await git_service.synchronize_project()

    assert any(["contains changes" in record.message for record in caplog.records])


@pytest.mark.parametrize(
    "repository_info",
    [
        {
            "name": "my-repo",
            "repository_url": "my-url",
            "ssh_key": "safe-key",
            "git_service": "Github",
            "git_service_access_token": "token",
            "target_branch": DEFAULT_TARGET_BRANCH,
        },
        {"repository_url": "my-url"},
        {"repository_url": "my-url2", "target_branch": "other"},
    ],
)
def test_save_repository_information(
    git_service: GitService, repository_info: Dict[Text, Text], metrics_events: List
):

    repository = RASA_REMOTE_REPOSITORY.copy()
    repository.update(repository_info)
    saved = git_service.save_repository(repository_info)
    assert saved == git_service.get_repository()

    actual = git_service.get_most_recent_repository().as_dict()
    # The database should have assigned an id
    assert actual.pop("id")

    # This is sensitive information and should not be returned
    fields_which_should_not_be_returned = {"ssh_key", "git_service_access_token"}
    assert all([field not in actual for field in fields_which_should_not_be_returned])

    # Let's check if the other fields are correct
    fields_to_check = repository_info.keys() - fields_which_should_not_be_returned
    assert all(
        [actual.get(field) == repository_info.get(field) for field in fields_to_check]
    )

    # Assert repository creation was tracked
    assert metrics_events


def test_save_repository_information_with_insufficient_permissions(
    git_service_with_credentials_validation: GitService, metrics_events: List
):
    invalid_repository_info = {"repository_url": "weird url"}
    with pytest.raises(ValueError):
        git_service_with_credentials_validation.save_repository(invalid_repository_info)

    # Assert nothing was tracked
    assert not metrics_events


def test_save_ssh_key(git_service: GitService):
    repository = RASA_REMOTE_REPOSITORY.copy()
    if not repository["ssh_key"]:
        # Only set the ssh key on travis since we can use the developer's key locally
        repository["ssh_key"] = 123

    saved = git_service.save_repository(repository)
    ssh_files_directory = git_service._directory_for_git_clones / SSH_FILES_DIRECTORY
    path_to_command = ssh_files_directory / f"{saved['id']}.sh"

    assert path_to_command.exists()
    assert oct(stat.S_IMODE(path_to_command.lstat().st_mode)) == "0o700"

    path_to_key = ssh_files_directory / f"{saved['id']}.key"

    assert path_to_key.exists()


def test_update_repository(git_service: GitService):
    old = {"repository_url": "my-test-url"}
    _ = git_service.save_repository(old)

    number_of_saved = len(git_service.get_repositories())
    updated = {
        "name": "my-repo",
        "repository_url": "my-url",
        "ssh_key": "safe-key",
        "git_service": "Github",
        "git_service_access_token": "token",
        "target_branch": DEFAULT_TARGET_BRANCH,
    }

    git_service.update_repository(updated)
    # There should be no new entry
    assert len(git_service.get_repositories()) == number_of_saved

    actual = git_service.get_repository()
    fields_to_check = updated.keys() - {"ssh_key", "git_service_access_token"}
    assert all([actual[field] == updated[field] for field in fields_to_check])


def test_delete_repository(git_service: GitService):
    saved = git_service.save_repository({"repository_url": "my-url"})
    git_service.delete_repository()

    repositories = git_service.get_repositories()
    assert all([repository["id"] != saved["id"] for repository in repositories])
    assert not git_service.repository_path().exists()


def test_delete_repository_if_id_does_not_exist(git_service: GitService):
    git_service.set_repository(9990)
    with pytest.raises(ValueError):
        git_service.delete_repository()


def test_get_status(git_service: GitService):
    result = git_service.get_repository_status()

    assert result.pop("time_of_last_pull") > 0
    assert result == {
        "is_committing_to_target_branch_allowed": True,
        "is_remote_ahead": False,
        "are_there_local_changes": False,
        "current_branch": DEFAULT_TARGET_BRANCH,
    }


def test_time_of_last_pull(git_service: GitService):
    old_pull_time = git_service._get_latest_fetch_time()

    git_service.merge_latest_remote_branch()

    assert git_service._get_latest_fetch_time() > old_pull_time


def test_has_required_remote_permission(
    git_service_with_credentials_validation: GitService,
):
    repository_url = RASA_REMOTE_REPOSITORY["repository_url"]

    assert git_service_with_credentials_validation.has_required_remote_permission(
        repository_url, SSH_KEY
    )


def test_has_required_remote_permission_if_not_the_case(
    git_service_with_credentials_validation: GitService,
):
    repository_url = "git@github.com:tmbo/questionary.git"

    assert not git_service_with_credentials_validation.has_required_remote_permission(
        repository_url, SSH_KEY
    )
